#
# VHA6
#
# Copyright (c) 2019 Varnish Software AS
#
# VHA_BROADCAST (SEND)
# - Detect cache insertion on any node
# - Broadcast the hash to the peers
# - This node is now the origin
#

acl vha6_localhost_origin {
	"127.0.0.0"/8;
	"::1"/128;
}

sub vcl_vha_internal {
	if (bereq.method ~ "^VHA" || bereq.http.vha6-method) {
		# Do nothing
		return (ok);
	}

	vha6_stats.counter("broadcast_candidates", 1, varnishstat = true,
		comment = "Broadcast candidates");

	if (bereq.uncacheable || beresp.uncacheable) {
		vha.log("VHA_BROADCAST ORIGIN: object uncacheable, skipping");
		vha6_stats.counter("broadcast_nocache", 1, varnishstat = true,
			comment = "Broadcast candidate uncacheable");
		return (ok);
	} else if (vha6_request.get("skip") == "true") {
		vha.log("VHA_BROADCAST ORIGIN: request override, skipping");
		vha6_stats.counter("broadcast_skip", 1, varnishstat = true,
			comment = "Broadcast candidate has a VCL override");
		return (ok);
	} else if (beresp.ttl <= std.duration(vha6_opts.get("min_ttl"), 3s)) {
		vha.log("VHA_BROADCAST ORIGIN: low ttl, skipping");
		vha6_stats.counter("broadcast_lowttl", 1, varnishstat = true,
			comment = "Broadcast candidate has a low TTL");
		return (ok);
	} else if (std.integer(beresp.http.Content-Length, 0) >
			std.integer(vha6_opts.get("max_bytes"), 25000000)) {
		vha.log("VHA_BROADCAST ORIGIN: too large, skipping");
		vha6_stats.counter("broadcast_toolarge", 1, varnishstat = true,
			comment = "Broadcast candidate is too large");
		return (ok);
	} else if (vha6_stats.counter("broadcasts_sec", 1, 1s) >
			std.integer(vha6_opts.get_any("broadcast_rate_limit,max_requests_sec", "200"), 0)) {
		vha.log("VHA_BROADCAST ORIGIN: max broadcasts/sec, skipping");
		vha6_stats.counter("error_rate_limited", 1, varnishstat = true,
			comment = "Rate limit hit");
		return (ok);
	} else if (http.task_count(9) >
			std.integer(vha6_opts.get_any("broadcast_limit,broadcast_rate_limit,max_requests_sec", "200"), 0)) {
		vha.log("VHA_BROADCAST ORIGIN: max active broadcasts, skipping");
		vha6_stats.counter("error_max_broadcasts", 1, varnishstat = true,
			comment = "Max active broadcasts hit");
		return (ok);
	}

	vha.log("VHA_BROADCAST ORIGIN: object candidate");

	# Setup hash broadcast
	http.init(9, log_level = MEDIUM);

	if (vha6_opts.get("broadcaster_ssl_verify_peer") == "false") {
		http.req_set_iparam(9,"CURLOPT_SSL_VERIFYPEER",0);
	}

	if (vha6_opts.get("broadcaster_ssl_verify_host") == "false") {
		http.req_set_iparam(9,"CURLOPT_SSL_VERIFYHOST",0);
	}

	http.req_set_url(9,
		vha6_opts.get("broadcaster_scheme", "http") + "://" +
		vha6_opts.get("broadcaster_host", "localhost") + ":" +
		vha6_opts.get("broadcaster_port", "8088") + "/" +
		blob.encode(HEX, LOWER, bereq.hash));
	http.req_set_method(9, "VHA_BROADCAST");
	http.req_set_header(9, "Host", bereq.http.Host);
	http.req_set_header(9, "vha6-url", bereq.url);
	http.req_set_header(9, "vha6-version", vha6_opts.get("version"));
	http.req_set_header(9, "vha6-xid", bereq.xid);
	http.req_set_header(9, "vha6-date", now);
	http.req_unset_header(9, "Accept");
	if (beresp.http.Vary) {
		http.req_set_header(9, "vha6-vary", beresp.http.Vary);
		http.req_copy_headers(9, beresp.http.Vary);
	}
	if (vha6_opts.get("force_update") == "true" ||
			vha6_request.get("force_update") == "true") {
		http.req_set_header(9, "vha6-force", "true");
	}

	# Broadcaster setup
	http.req_set_header(9, "X-Broadcast-Skip", server.identity);
	http.req_unset_header(9, "X-Broadcast-Group");
	if (vha6_opts.get("broadcaster_group")) {
		http.req_set_header(9, "X-Broadcast-Group", vha6_opts.get("broadcaster_group"));
	}

	# Figure out origin address
	if (vha6_opts.get("origin")) {
		vha.log("VHA_BROADCAST ORIGIN ADDR: origin override");
		http.req_set_header(9, "vha6-origin",
			vha6_opts.get("origin_scheme", "http") + "://" +
			vha6_opts.get("origin") + ":" +
			vha6_opts.get("origin_port", std.port(server.ip)));
	} else if (vha6_nodes.get(vha6_opts.get("broadcaster_group") + "_" + server.identity)) {
		vha.log("VHA_BROADCAST ORIGIN ADDR: nodes.conf group");
		http.req_set_header(9, "vha6-origin",
			vha6_nodes.get(vha6_opts.get("broadcaster_group") + "_" + server.identity));
	} else if (vha6_nodes.get(server.identity)) {
		vha.log("VHA_BROADCAST ORIGIN ADDR: nodes.conf");
		http.req_set_header(9, "vha6-origin", vha6_nodes.get(server.identity));
	} else if ((vha6_opts.get("allow_localhost") != "true" &&
			# Alias to fix spelling mistake locahost
			vha6_opts.get("allow_locahost") != "true" ) &&
			server.ip ~ vha6_localhost_origin) {
		vha.log("VHA_BROADCAST ORIGIN: localhost origin detected, skipping");
		vha6_stats.counter("error_localhost", 1, varnishstat = true,
			comment = "VHA origin points back to localhost");
		return (ok);
	} else if (std.port(server.ip) == 443) {
		vha.log("VHA_BROADCAST ORIGIN ADDR: server.ip https");
		http.req_set_header(9, "vha6-origin",
			vha6_opts.get("origin_scheme", "https") + "://" +
			server.ip + ":" +
			vha6_opts.get("origin_port", std.port(server.ip)));
	} else {
		vha.log("VHA_BROADCAST ORIGIN ADDR: server.ip");
		http.req_set_header(9, "vha6-origin",
			vha6_opts.get("origin_scheme", "http") + "://" +
			server.ip + ":" +
			vha6_opts.get("origin_port", std.port(server.ip)));
	}
	http.req_set_header(9, "vha6-origin-id", server.identity);

	# Export callbacks
	call fast_304_broadcast_insert;
	call edgestash_broadcast_insert;

	# Write HMAC
	call vha_token_broadcast;

	# Did we abort during this broadcast?
	if (vha6_request.get("skip") == "true") {
		vha.log("VHA_BROADCAST ORIGIN: request override during broadcast, skipping");
		vha6_stats.counter("broadcast_skip", 1, varnishstat = true,
			comment = "Broadcast candidate has a VCL override");
		return (ok);
	}

	# Hook
	call vha6_hook_broadcast;

	# Send broadcast
	if (vha6_opts.get("__broadcaster_async") == "false") {
		http.req_send(9);
		http.resp_wait(9);
		vha.log("VHA_BROADCAST ASYNC: " + http.resp_get_errorcode(9) + " " +
			http.resp_get_status(9) + " " + http.resp_get_reason(9));
		http.finish(9);
	} else {
		http.req_send_and_finish(9, (vha6_opts.get("broadcaster_force_sync") == "true"));
	}

	vha6_stats.counter("broadcasts", 1, varnishstat = true,
		comment = "Successful broadcasts");
	if (vha6_opts.get("vcs", "true") == "true") {
		std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/" + server.identity + "/broadcasts");
		std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/broadcasts");

	}
	vha6_request.set("skip", "true");
}

# EOF
