#
# VHA6
#
# Copyright (c) 2019 Varnish Software AS
#
# VHA_FETCH (RECEIVE)
# - Receive the VHA_FETCH from the peer(s)
# - Lookup the hash
# - Return the object with stevedore attributes to the peer
#

sub vcl_recv {
	if (req.method == "VHA_FETCH") {
		# Hook
		call vha6_hook_origin_recv;

		# Security check
		call vha_security_check;

		# Assert we are correct
		if (req.http.vha6-origin-id != server.identity) {
			vha.log("VHA_FETCH ORIGIN ERROR: origin id mismatch");
			vha6_stats.counter("error_origin_mismatch", 1, varnishstat = true,
				comment = "Origin received the wrong fetch");
			return (synth(403, "NODE ID"));
		}

		# Extend keep-alive
		set sess.timeout_idle = std.duration(vha6_opts.get("keep_alive"), 120s);

		# Do not serve a stale object
		if (vha6_opts.get("allow_stale") != "true") {
			set req.grace = 0s;
		}

		# Do not attempt ban evaluation
		vha.skip_ban();

		vha.log("VHA_FETCH ORIGIN");
		vha6_stats.counter("fetch_origin", 1, varnishstat = true,
			comment = "Fetches which hit this origin node");
		set req.http.vha6-method = "VHA_FETCH";

		return (hash);
	}
}

sub vcl_hash {
	if (req.method == "VHA_FETCH") {
		set req.hash = blob.decode(HEX, 0, regsub(req.url, "^\/", ""));
		return (lookup);
	}
}

sub vcl_hit {
	if (req.method == "VHA_FETCH") {
		return (deliver);
	}
}

sub vcl_miss {
	if (req.method == "VHA_FETCH") {
		vha.log("VHA_FETCH ORIGIN ERROR: cache miss");
		vha6_stats.counter("error_origin_miss", 1, varnishstat = true,
			comment = "Origin is missing the object");
		return (synth(500, "ORIGIN MISS"));
	}
}

sub vcl_deliver {
	if (req.method == "VHA_FETCH") {
		vha.log("VHA_FETCH TRANSFER");

		set resp.http.vha6-origin = server.identity;
		vha6_stats.counter("fetch_origin_deliver", 1, varnishstat = true,
			comment = "Fetches which were delivered from origin to the peer");
		if (vha6_opts.get("vcs", "true") == "true") {
			std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/" + server.identity + "/fetch_origin_deliver");
			std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/fetch_origin_deliver");
		}

		# Security
		call vha_seal_calculate;

		# Hook
		call vha6_hook_origin_deliver;

		return (deliver);
	}
}

# EOF
