#
# VHA6
#
# Copyright (c) 2019 Varnish Software AS
#
# VHA_BROADCAST (RECEIVE)
# VHA_FETCH (SEND)
# - Receive the VHA_BROADCAST from the origin
# - This node is now a peer
# - Lookup the hash
# - VHA_FETCH the object back from the origin
# - Insert object and stevedore attributes
#

sub vcl_recv {
	if (req.method == "VHA_BROADCAST") {
		# Hook
		call vha6_hook_peer_recv;

		# Security
		call vha_security_check;

		# Detect loop
		if (req.http.vha6-origin-id == server.identity) {
			vha.log("VHA_BROADCAST PEER: self");
			vha6_stats.counter("error_fetch_self", 1, varnishstat = true,
				comment = "Broadcasts which hit the origin node");
			return (synth(200, "SELF"));
		}

		# Extend keep-alive
		set sess.timeout_idle = std.duration(vha6_opts.get("keep_alive"), 120s);

		# Force update?
		if (req.http.vha6-force == "true" || vha6_opts.get("force_update") == "true") {
			set req.hash_always_miss = true;
		}

		# Make sure we update a stale object
		set req.grace = 0s;

		# Do not attempt ban evaluation
		vha.skip_ban();

		vha.log("VHA_BROADCAST PEER");
		vha6_stats.counter("fetch_peer", 1, varnishstat = true,
			comment = "Broadcasts which hit this peer node (fetches)");
		set req.http.vha6-method = "VHA_BROADCAST";
		set req.http.vha6-peer-id = server.identity;

		return (hash);
	}
}

sub vcl_hash {
	if (req.method == "VHA_BROADCAST") {
		set req.hash = blob.decode(HEX, 0, regsub(req.url, "^\/", ""));
		return (lookup);
	}
}

sub vcl_miss {
	if (req.method == "VHA_BROADCAST") {
		return (fetch);
	}
}

sub vcl_hit {
	if (req.method == "VHA_BROADCAST") {
		vha.log("VHA_BROADCAST PEER: cache hit");
		vha6_stats.counter("fetch_peer_hit", 1, varnishstat = true,
			comment = "Broadcasts which cache hit on the peer node");
		return (synth(200, "PEER HIT"));
	}
}

sub vcl_backend_fetch {
	if (bereq.http.vha6-method == "VHA_BROADCAST") {
		set bereq.method = "VHA_FETCH";
		set bereq.backend = goto.dns_backend(
			bereq.http.vha6-origin,
			host_header = bereq.http.Host,
			ttl = std.duration(vha6_opts.get("origin_backend_linger"), 10s),
			ssl = (vha6_opts.get("origin_ssl") == "true"),
			ssl_sni = (vha6_opts.get("origin_ssl_sni", "true") == "true"),
			ssl_verify_peer = (vha6_opts.get("origin_ssl_verify_peer", "true") == "true"),
			ssl_verify_host = (vha6_opts.get("origin_ssl_verify_host") == "true"));

		if (std.duration(vha6_opts.get("fetch_timeout"), 0s) > 0s) {
			set bereq.first_byte_timeout = std.duration(vha6_opts.get("fetch_timeout"), 0s);
		}

		# Hook
		call vha6_hook_peer_fetch;

		vha.log("VHA_BROADCAST PEER: fetch");

		return (fetch);
	}
}

sub vcl_backend_response {
	if (bereq.method == "VHA_FETCH") {
		if (beresp.http.vha6-error) {
			vha.log("VHA_BROADCAST PEER ERROR: " + beresp.http.vha6-error);
			vha6_stats.counter("error_fetch_insert", 1, varnishstat = true,
				comment = "Fetches which encoutnered an origin error");
			return (abandon);
		}

		# Hook
		call vha6_hook_peer_insert;

		# Security
		call vha_seal_validate;

		# Disable streaming on peer insert
		if (vha6_opts.get("peer_stream") == "false") {
			set beresp.do_stream = false;
		}

		vha.log("VHA_BROADCAST INSERT");
		vha6_stats.counter("fetch_peer_insert", 1, varnishstat = true,
			comment = "Fetches which were successfully inserted");
		if (vha6_opts.get("vcs", "true") == "true") {
			std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/" + server.identity + "/fetch_peer_insert");
			std.log(vha6_opts.get("vcs_key", "vcs-key") + ":VHA6/fetch_peer_insert");
		}
		vha6_request.set("skip", "true");

		return (deliver);
	}
}

sub vcl_backend_error {
	if (bereq.method == "VHA_FETCH") {
		vha.log("VHA_BROADCAST PEER ERROR: backend_error");
		vha6_stats.counter("error_fetch", 1, varnishstat = true,
			comment = "Fetches which encountered a network error");
		return (abandon);
	}
}

sub vcl_deliver {
	if (req.method == "VHA_BROADCAST") {
		vha.log("VHA_BROADCAST PEER: done");
		return (synth(200, "DONE"));
	}
}

# EOF
