#
# VHA6
#
# Copyright (c) 2019 Varnish Software AS
#
# Security for VHA6 commands
#

sub vha6_token_init {
	if (vha6_opts.get("token", "") == "") {
		return (fail({"VHA6 token not found, hint: vha6_opts.set("token", "secret");"}));
	}

	new vha6_hmac = crypto.hmac_init(sha256);
	vha6_hmac.set_key(crypto.hash(sha256, vha6_opts.get("token")));
	vha6_opts.delete("token");
}

# Generate an token HMAC
sub vha_token_broadcast {
	vha6_hmac.add_string("hash/" + blob.encode(HEX, LOWER, bereq.hash));
	vha6_hmac.add_string("url" + http.req_get_header(9, "vha6-url"));
	vha6_hmac.add_string("version" + http.req_get_header(9, "vha6-version"));
	vha6_hmac.add_string("date" + http.req_get_header(9, "vha6-date"));
	vha6_hmac.add_string("origin" + http.req_get_header(9, "vha6-origin"));
	vha6_hmac.add_string("origin-id" + http.req_get_header(9, "vha6-origin-id"));

	http.req_set_header(9, "vha6-token", crypto.hex_encode(vha6_hmac.get()));
}

# Calculate the matching token HMAC for verification
sub vha_token_calculate {
	if (req.method != "VHA_FETCH" && req.method != "VHA_BROADCAST") {
		return (synth(500, "INVALID"));
	}

	vha6_hmac.add_string("hash" + req.url);
	vha6_hmac.add_string("url" + req.http.vha6-url);
	vha6_hmac.add_string("version" + req.http.vha6-version);
	vha6_hmac.add_string("date" + req.http.vha6-date);
	vha6_hmac.add_string("origin" + req.http.vha6-origin);
	vha6_hmac.add_string("origin-id" + req.http.vha6-origin-id);
}

sub vha_security_check {
	if (req.method != "VHA_FETCH" && req.method != "VHA_BROADCAST") {
		return (synth(500, "INVALID"));
	}

	# Version check
	if (vha6_opts.get("version") != req.http.vha6-version) {
		vha.log(req.method + " PEER ERROR: version mismatch");
		vha6_stats.counter("error_version_mismatch", 1, varnishstat = true,
			comment = "VHA version mismatch");
		return (synth(403, "VERSION"));
	}

	# Token valid?
	if (!req.http.vha6-token || req.http.vha6-token == "") {
		vha.log(req.method + " PEER ERROR: bad token");
		vha6_stats.counter("error_no_token", 1, varnishstat = true,
			comment = "No token found");
		return (synth(403, "TOKEN"));
	}

	# Token HMAC check
	call vha_token_calculate;
	if (crypto.hex_encode(vha6_hmac.get()) != req.http.vha6-token) {
		vha.log(req.method + " PEER ERROR: bad token");
		vha6_stats.counter("error_bad_token", 1, varnishstat = true,
			comment = "Tokens do not match");
		return (synth(403, "TOKEN"));
	}

	# Token TTL check
	if (std.time(req.http.vha6-date, now - 100y) +
			std.duration(vha6_opts.get("token_ttl"), 2m) < now) {
		vha.log(req.method + " ORIGIN ERROR: token expired");
		vha6_stats.counter("error_stale_token", 1, varnishstat = true,
			comment = "Token has expired");
		return (synth(403, "TOKEN TTL"));
	}

	# Rate limit
	if (vha6_stats.counter("requests_sec", 1, 1s) >
			std.integer(vha6_opts.get("request_rate_limit", ""),
			std.integer(vha6_opts.get_any("broadcast_rate_limit,max_requests_sec", ""), 200) * 5 / 2)) {
		vha.log(req.method + " PEER: max requests/sec");
		vha6_stats.counter("error_rate_limited", 1, varnishstat = true,
			comment = "Rate limit hit");
		return (synth(403, "RATE"));
	}
}

# Calculate an origin seal HMAC
sub vha_seal_calculate {
	if (req.method != "VHA_FETCH" && req.method != "VHA_BROADCAST") {
		return (synth(500, "INVALID"));
	}

	# Calculate seal HMAC
	vha6_hmac.add_string("seal" + req.http.vha6-token);

	set resp.http.vha6-seal = crypto.hex_encode(vha6_hmac.get());
}

# Validate an origin seal HMAC
sub vha_seal_validate {
	if (bereq.method != "VHA_FETCH") {
		return (abandon);
	}

	# Calculate seal HMAC
	vha6_hmac.add_string("seal" + bereq.http.vha6-token);

	if (beresp.http.vha6-seal != crypto.hex_encode(vha6_hmac.get())) {
		set beresp.http.vha6-error = "BAD SEAL";
		vha.log("VHA_BROADCAST PEER BAD SEAL");
		vha6_stats.counter("error_fetch_seal", 1, varnishstat = true,
			comment = "Invalid origin response");
		return (abandon);
	}

	unset beresp.http.vha6-seal;
}

# EOF
