# To use this vcl, simply include it at the top of your vcl, all
# necessary subroutines will be automatically called.
# requests with a body will be replicated if std.cache_req_body() was used.
import goto;
import kvstore;
import std;
import vha;

/*
 *
 *
 * In your vcl_init, set keys of vha_opts to secure replication, for example:
 *
 *     sub vcl_init {
 *         vha_opts.set("token", "example_token123");
 *     }
 *
 * For ease of use, all variables are strings, here are the available ones:
 * - token: replication requests must carry this token to be authorized
 * - port: replication requests must arrive on this port to be authorized
 * - ip: replication requests must be made to that ip to be authorized
 * - accept: default to "secure", meaning requests will be tested
 *   against token, port and/or ip (at least one must be set). "all" deactivates
 *   security checks and "none" rejects all replication requests.
 * - no-callback: don't use the vha-origin to loop the request back to the
 *   other vha node, go to the regular backend instead. If the value is set,
 *   the option is active.
 * - replicate_body: the body content is given to VHA via the log, meaning that
 *   sensitive data may appear here. By default, the option is unset and won't
 *   do anything unless std.cache_req_body() is called.
 */
sub vcl_init			{ call vha_init;		}
sub vcl_recv			{ call vha_recv;		}
sub vcl_hit			{ call vha_hit;			}
sub vcl_miss			{ call vha_miss;		}
sub vcl_pass			{ call vha_pass;		}
sub vcl_pipe			{ call vha_pipe;		}
sub vcl_deliver			{ call vha_deliver;		}
sub vcl_backend_fetch		{ call vha_backend_fetch;	}
sub vcl_backend_response	{ call vha_backend_response;	}
sub vcl_backend_error		{ call vha_backend_error;	}

# initialize the option store
sub vha_init {
	new vha_opts = kvstore.init(16);
}

sub vha_recv_failedauth {
	set req.http.vha-error = "Authorization failed";
	return (synth(403));
}

# Refuse the request if there's a vha-origin server but the condition isn't
# satisfied.
sub vha_recv {
	# forbid replication if vsl_reclen is too short for this request
	vha.check();
	# Replication errors have at least a vha-origin header
	if ((req.http.vha-origin || req.http.vha-fetch || req.http.vha-status)
	    && vha_opts.get("accept") != "all") {
		if (req.http.vha-version != "3.0") {
			set req.http.vha-error = "Wrong VHA version";
			return (synth(400));
		}
		if (!req.http.vha-origin) {
			call vha_recv_failedauth;
		}
		if (vha_opts.get("accept") == "none") {
			call vha_recv_failedauth;
		}
		if (!vha_opts.get("token") &&
		    !vha_opts.get("ip") &&
		    !vha_opts.get("port")) {
			call vha_recv_failedauth;
		}
		if (vha_opts.get("token") &&
		    vha_opts.get("token") != req.http.vha-token) {
			call vha_recv_failedauth;
		}
		if (vha_opts.get("ip")) {
			set req.http.vha-ip = server.ip;
			if (req.http.vha-ip != vha_opts.get("ip")) {
				unset req.http.vha-ip;
				call vha_recv_failedauth;
			}
			unset req.http.vha-ip;
		}
		if (vha_opts.get("port")) {
			set req.http.vha-port = std.port(server.ip);
			if (req.http.vha-port != vha_opts.get("port")) {
				unset req.http.vha-port;
				call vha_recv_failedauth;
			}
			unset req.http.vha-port;
		}
	}

	# explain the source of the HTTP request
	if (req.http.vha-origin) {
		if (req.http.vha-fetch) {
			set req.http.vha-status = "from-varnish";
		} else {
			set req.http.vha-status = "from-vha";
		}
	}
}

# don't deliver expired objects, as it could trigger hits-for-passes
sub vha_hit {
	if (req.http.vha-origin && req.http.vha-fetch && obj.ttl < 1s) {
		return (miss);
	}
}

# reset the request
sub vha_miss {
	if (req.http.vha-origin && !req.http.vha-fetch) {
		vha.rollback();
	}
}

# flag the request as non-replicable, preventing the body dump.
sub vha_pass {
	vha.forbid_replication("request was a pass");
}

# avoid piping replication request
sub vha_pipe {
	if (req.http.vha-origin) {
		return (synth (503));
	}
}

# Avoid header explosion. VHA-enabled nodes will revalidate content,
# bouncing it back-and-forth, accumulating headers, so we unset them.
sub vha_deliver {
	# if vha-origin is present, we are replying to a fellow node
	# and need to unset esi
	if (req.http.vha-origin) {
		set req.esi = false;
		unset resp.http.X-Varnish;
		unset resp.http.Via;
	} elsif (req.method != "HEAD" &&
	    (req.http.Content-Type || req.http.Content-Length) &&
	    vha.replication_allowed()) {
		if (vha_opts.get("replicate_body")) {
			# replicate body. This is done as late as possible
			# to make sure we really have to do it
			vha.replicate_body();
		}
	}
}

# Override the backend in case of replication then return(fetch).
sub vha_backend_fetch {
	if (bereq.http.vha-origin) {
		if (bereq.http.vha-fetch) {
			# this request comes from another varnish (the backend
			# was already overridden), abort
			set bereq.http.vha-error = "backend already overridden";
			# set an NULL backend so that we go to vcl_backend_error
			set bereq.backend = vha.fail_backend();
		} else {
			if (!vha_opts.get("no-callback")) {
				# we grab the callback IP and set vha-fetch
				set bereq.http.vha-fetch = regsub(bereq.http.vha-origin, "^([^ ]+).*", "\1");
				set bereq.backend = goto.dns_backend(bereq.http.vha-fetch,
				    host_header = bereq.http.host);
			}
			if (bereq.http.vha-method) {
				set bereq.method = bereq.http.vha-method;
			}
		}
		return (fetch);
	}
}

# Make sure to not cache failed replication requests
sub vha_backend_response {
	if (bereq.http.vha-origin && beresp.http.vha-error) {
		return (abandon);
	}
}

# Flag the faulty fetch as a vha error
sub vha_backend_error {
	if (bereq.http.vha-origin && bereq.http.vha-error) {
		set beresp.http.vha-error = bereq.http.vha-error;
		return (deliver);
	}
}
